
/*!
\file terralib/mnt/core/Utils.h

\brief Utility functions for MNT support.
*/

#ifndef __TERRALIB_MNT_INTERNAL_UTILS_H
#define __TERRALIB_MNT_INTERNAL_UTILS_H

#include "Config.h"
#include "Enums.h"

#include "../../common/UnitsOfMeasureManager.h"

#include "../../dataaccess/dataset/DataSet.h"
#include "../../dataaccess/dataset/DataSetType.h"
#include "../../dataaccess/datasource/DataSource.h"

#include "../../geometry/Coord2D.h"
#include "../../geometry/LineString.h"
#include "../../geometry/MultiLineString.h"
#include "../../geometry/MultiPoint.h"
#include "../../geometry/Point.h"


#ifndef MIN
#define MIN(x, y) ( ((x) < (y)) ? (x) : (y) )
#endif

#ifndef MAX
#define MAX(x, y) ( ((x) > (y)) ? (x) : (y) )
#endif


namespace te
{
  namespace mnt
  {
    size_t ReadPoints(  std::string &inDsetName, te::da::DataSourcePtr &inDsrc, std::string &atrZ, double tol, te::gm::MultiPoint &mpt, std::string &geostype, 
                        te::gm::Envelope &env);

    size_t ReadSamples( std::string &inDsetName, te::da::DataSourcePtr &inDsrc, std::string &atrZ, double tol, double max, Simplify alg, te::gm::MultiPoint &mpt,
                        te::gm::MultiLineString &isolines, std::string &geostype, te::gm::Envelope &env, int srid = 0);

    double Distance(const te::gm::Coord2D &pt1, const te::gm::Coord2D &pt2);

    double Distance(const double pt1x, const double pt1y, const double pt2x, const double pt2y);

    bool Equal(te::gm::Point &p1, te::gm::Point &p2, double &tol);

    //DouglasPeucker from TerraAmazon 4
    te::gm::LineString* DouglasPeuckerTA(te::gm::LineString *lineIn, double simpFactor, double Zvalue);

    //DouglasPeucker from libgeos
    te::gm::LineString* GEOS_DouglasPeucker(te::gm::LineString *ls, double snap, double Zvalue);

    TEMNTEXPORT te::gm::LineString* pointListSimplify(te::gm::LineString *ls, double snap, double maxdist, double Zvalue);

    double SegmentDistance(double fx, double fy, double lx, double ly, double ptx, double pty, double *pix, double *piy);
      
    double pointToSegmentDistance(te::gm::Point &fseg, te::gm::Point &lseg, te::gm::Point &pt, te::gm::Point *pti);

    double coordToSegmentDistance(te::gm::Coord2D &fseg, te::gm::Coord2D &lseg, te::gm::Coord2D &pt, te::gm::Coord2D *pti);

    // Find center point of triangle using its vertices.
    short findCenter(te::gm::Point* vert, double* pcx, double* pcy);

    short segIntersect(te::gm::Point &pfr, te::gm::Point &pto, te::gm::Point &lfr, te::gm::Point &lto);
    bool segInterPoint(te::gm::Point &pfr, te::gm::Point &pto, te::gm::Point &lfr, te::gm::Point &lto, te::gm::Point *pt);

    // Test Vertex values for valid and special cases, returns TRUE if normal, FALSE otherwise.
    bool testVertexValues(double, te::gm::Point*);

    // Define intersections of contour with edges of triangle
    bool defineInterTriangle(std::vector<te::gm::Point>&, te::gm::Point*, double );

    // Define intersection point of contour with edge between two points
    bool defineInterEdge(std::vector<te::gm::Point>&, te::gm::Point&, te::gm::Point&, double);

    //Extract lines from pline and store connected ones in clinlist.
    bool extractLines(std::vector<te::gm::Point> &pline, std::vector<te::gm::LineString> &clinlist, double scale);

    bool initLineVector(std::vector<te::gm::Point> &pline, std::vector<te::gm::Point> &vect);
    bool assembLine(std::vector<te::gm::LineString> &linlout, std::vector<te::gm::Point> &vect);

    bool equalFptSpt(te::gm::Point & fpt, te::gm::Point &spt, double scale);

    te::gm::Point Max(te::gm::Point& p1, te::gm::Point& p2);
    te::gm::Point Min(te::gm::Point& p1, te::gm::Point& p2);

    // Calculate Normal vector of Triangle
    bool triangleNormalVector(te::gm::Point*, double*);

    bool normalizeVector(double*);

    bool onSegment(te::gm::Point& pt, te::gm::Point& fseg, te::gm::Point& lseg, double tol);

    int onSameSide(te::gm::Coord2D pt1, te::gm::Coord2D pt2, te::gm::Coord2D fseg, te::gm::Coord2D lseg);

    bool point3dListFilter(std::vector<te::gm::Point> &p3dl, std::vector<bool> &fixed, double tol);

    bool Least_square_fitting(double *vectx, double *vecty, short np, short deg, double *coef);

    bool Gauss_elimination(short m, short n, double mat[6][6]);

    bool SaveIso(std::string& outDsetName, te::da::DataSourcePtr &outDsrc, std::vector<te::gm::LineString> &isolist, std::vector<double> &guidevalues, int srid);

    void Save(te::da::DataSource* source, te::da::DataSet* result, te::da::DataSetType* outDsType);

    TEMNTEXPORT bool convertAngleToPlanar(double& val, te::common::UnitOfMeasurePtr planar, int type);
    TEMNTEXPORT bool convertPlanarToAngle(double& val, te::common::UnitOfMeasurePtr unit);

    TEMNTEXPORT void getMinMax(te::rst::Raster* inputRst, double &vmin, double &vmax);
 
    double PerpendicularDistance(te::gm::Coord2D& first, te::gm::Coord2D& last, te::gm::Coord2D& pin, te::gm::Coord2D& pinter);

    TEMNTEXPORT te::mnt::mntType getMNTType(const te::da::DataSetType* dt);
  }
}

#endif
