/*
  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

  This file is part of the TerraLib - a Framework for building GIS enabled
  applications.

  TerraLib is free software: you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation, either version 3 of the License,
  or (at your option) any later version.

  TerraLib is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with TerraLib. See COPYING. If not, write to
  TerraLib Team at <terralib-team@terralib.org>.
*/

/*!
  \file terralib/qt/plugins/wtss/WtssDialog.cpp

  \brief Configuration dialog for Web Time Series Services plugin.

  \author Matheus Cavassan Zaglia
 */

// TerraLib
#include "../../../maptools/MarkRendererManager.h"
#include "../../../se/Fill.h"
#include "../../../se/Mark.h"
#include "../../../se/Stroke.h"
#include "../../../se/Utils.h"
#include "../../widgets/tools/AbstractTool.h"
#include "../../widgets/tools/PointPicker.h"
#include "../../widgets/utils/ScopedCursor.h"

// QT
#include <QFileDialog>
#include <QInputDialog>
#include <QMessageBox>

//Qwt
#include <qwt_legend.h>

// STL
#include <ctime>
#include <fstream>

// Boost
#include <boost/format.hpp>

// wtss.cxx
#include <wtss-cxx/wtss.hpp>

#include "WtssDialog.h"
#include "Exception.h"
#include "ServerManager.h"
#include "ui_WtssDialogForm.h"

namespace cxx = wtss::cxx;

te::qt::plugins::wtss::WtssDialog::WtssDialog(QWidget *parent,
                                                Qt::WindowFlags f)
    : QDialog(parent, f),
      m_ui(new Ui::WtssDialogForm),
      dirty(false),
      m_checkServer(0),
      m_checkCoverage(0),
      m_checkAttribute(0),
      m_mapDisplay(0),
      m_chartDisplay(0),
      m_chartStyle(0),
      m_timeSeriesChart(0),
      m_timeSeriesVec(0)
{
  m_ui->setupUi(this);

  this->setWindowTitle(tr("Web Time Series Services"));

  m_ui->btnAddServer->setIcon(QIcon::fromTheme("list-add"));
  m_ui->btnRemoveServer->setIcon(QIcon::fromTheme("list-remove"));
  m_ui->btnRefreshServer->setIcon(QIcon::fromTheme("view-refresh"));
  m_ui->m_hideToolButton->setIcon(QIcon::fromTheme("hide"));
  m_ui->m_hideCoordSToolButton->setIcon(QIcon::fromTheme("hide"));

  m_ui->startLabel->setVisible(false);
  m_ui->endLabel->setVisible(false);
  m_ui->m_startDateEdit->setVisible(false);
  m_ui->m_endDateEdit->setVisible(false);

  defineDisplay();

  defineMarker();

  srand(time(NULL));

  connect(m_ui->btnAddServer, SIGNAL(clicked()), this,
          SLOT(onServerAddButtonClicked()));
  connect(m_ui->btnRemoveServer, SIGNAL(clicked()), this,
          SLOT(onServerRemoveButtonClicked()));
  connect(m_ui->m_serverTreeWidget, SIGNAL(itemChanged(QTreeWidgetItem *, int)),
          this, SLOT(onItemChecked(QTreeWidgetItem *)));
  connect(m_ui->btnHelp, SIGNAL(clicked()), this, SLOT(onHelpButtonClicked()));
  connect(m_ui->btnClose, SIGNAL(clicked()), this, SLOT(close()));
  connect(m_ui->m_hideToolButton, SIGNAL(clicked()), this,
          SLOT(onHideButtonClicked()));
  connect(m_ui->btnRefreshServer, SIGNAL(clicked()), this,
          SLOT(onServerRefreshButtonClicked()));
  connect(m_ui->btnImport, SIGNAL(clicked()), this,
          SLOT(onImportGraphClicked()));
  connect(m_ui->btnExport, SIGNAL(clicked()), this,
          SLOT(onExportGraphClicked()));
  connect(m_ui->btnQuery, SIGNAL(clicked()), this,
          SLOT(onQueryButtonClicked()));
  connect(m_ui->m_hideCoordSToolButton, SIGNAL(clicked()), this,
          SLOT(onHideCoordSelectedsClicked()));
  connect(m_ui->m_coordSListWidget, SIGNAL(itemClicked(QListWidgetItem *)),
          this, SLOT(onAddCoordToList(QListWidgetItem *)));

  loadSettings();

  dirty = true;
}

te::qt::plugins::wtss::WtssDialog::~WtssDialog()
{
}

void te::qt::plugins::wtss::WtssDialog::setMapDisplay(
    te::qt::widgets::MapDisplay *mapDisplay)
{
  m_mapDisplay = mapDisplay;
}

void te::qt::plugins::wtss::WtssDialog::onPointPickerToggled(bool checked)
{
  if(!checked)
    return;

  te::qt::widgets::PointPicker *point =
      new te::qt::widgets::PointPicker(m_mapDisplay, Qt::CrossCursor);
  m_mapDisplay->setCurrentTool(point);

  connect(point, SIGNAL(pointPicked(QPointF &)), this,
          SLOT(onGetPointCoordinate(QPointF &)));
}

void te::qt::plugins::wtss::WtssDialog::onServerAddButtonClicked()
{
  dirty = false;

  QInputDialog *inputDialog = new QInputDialog();
  inputDialog->setOptions(QInputDialog::NoButtons);

  bool accept;

  QString uri = inputDialog->getText(NULL, "Add Server", "Server URI:",
                                     QLineEdit::Normal, "", &accept);
  if(!accept)
  {
    dirty = true;
    return;
  }

  if(!uri.isEmpty())
  {
    te::qt::widgets::ScopedCursor c(Qt::WaitCursor);

    try
    {
      te::qt::plugins::wtss::ServerManager::getInstance().addServer(uri);

      j_config = te::qt::plugins::wtss::ServerManager::getInstance()
                     .loadSettings()
                     .object();

      addServer(uri);
    }
    catch(...)
    {
      QMessageBox::warning(this, tr("Web Time Series"),
                           tr("An error has occurred, "
                              "please retype the wtss "
                              "server address."));
    }
  }
  else
  {
    QMessageBox::warning(this, tr("Web Time Series"),
                         tr("Please, input a server."));
  }

  dirty = true;
}

void te::qt::plugins::wtss::WtssDialog::onServerRemoveButtonClicked()
{
  dirty = false;

  if(m_ui->m_serverTreeWidget->currentItem() &&
     !m_ui->m_serverTreeWidget->currentItem()->parent())
  {
    QMessageBox::StandardButton reply = QMessageBox::question(
        this, "Remove server",
        "Remove '" + m_ui->m_serverTreeWidget->currentItem()->text(0) +
            "' server?",
        QMessageBox::Yes | QMessageBox::No);

    if(reply == QMessageBox::Yes)
    {
      try
      {
        te::qt::plugins::wtss::ServerManager::getInstance().removeServer(
            m_ui->m_serverTreeWidget->currentItem()->text(0));

        j_config = te::qt::plugins::wtss::ServerManager::getInstance()
                       .loadSettings()
                       .object();

        delete m_ui->m_serverTreeWidget->currentItem();

        m_chartDisplay->detachItems(te::qt::widgets::TIMESERIES_CHART);

        QMessageBox::information(this, tr("Web Time Series"),
                                 tr("The server was removed with success."));
      }
      catch(...)
      {
        QMessageBox::warning(this, tr("Web Time Series"),
                             tr("An error has occurred, please try again."));
      }
    }
  }
  else
  {
    QMessageBox::warning(this, tr("Web Time Series"),
                         tr("Please, select a server."));
  }

  dirty = true;
}

void te::qt::plugins::wtss::WtssDialog::onServerRefreshButtonClicked()
{
  dirty = false;
  if(m_ui->m_serverTreeWidget->currentItem() &&
     !m_ui->m_serverTreeWidget->currentItem()->parent())
  {
    try
    {
      QString uri = m_ui->m_serverTreeWidget->currentItem()->text(0);

      te::qt::plugins::wtss::ServerManager::getInstance().refreshServer(uri);

      delete m_ui->m_serverTreeWidget->currentItem();

      addServer(uri);
    }
    catch(...)
    {
      QMessageBox::warning(this, tr("Web Time Series"),
                           tr("An error has occurred, "
                              "please check server list."));
    }
  }
  else
  {
    QMessageBox::warning(this, tr("Web Time Series"),
                         tr("Please, select a server."));
  }

  dirty = true;
}

void te::qt::plugins::wtss::WtssDialog::onItemChecked(QTreeWidgetItem *item)
{
  if(dirty)
  {
    if(item->type() == treeItemType::attribute)
    {
      QString serverUri = item->parent()->parent()->text(0);
      QString cvName = item->parent()->text(0);
      QString attribute = item->text(0);

      te::qt::plugins::wtss::ServerManager::getInstance()
          .changeStatusAttribute(serverUri, cvName, attribute);
    }
    else if(item->type() == treeItemType::coverage)
    {
      QString serverUri = item->parent()->text(0);
      QString cvName = item->text(0);
      te::qt::plugins::wtss::ServerManager::getInstance().changeStatusCoverage(
          serverUri, cvName);
      dirty = false;
      for(int i = 0; i < item->parent()->childCount(); ++i)
      {
        QTreeWidgetItem *cv = item->parent()->child(i);
        if(cv->text(0) != cvName)
        {
          if(cv->checkState(0))
            cv->setCheckState(0, Qt::Unchecked);
        }
      }
      dirty = true;
    }
    else if(item->type() == treeItemType::server)
    {
      QString serverUri = item->text(0);

      te::qt::plugins::wtss::ServerManager::getInstance().changeStatusServer(
          serverUri);
      dirty = false;
      int size = m_ui->m_serverTreeWidget->topLevelItemCount();
      for(int i = 0; i < size; ++i)
      {
        QTreeWidgetItem *sv = m_ui->m_serverTreeWidget->topLevelItem(i);
        QString server = sv->text(0);
        if(server != serverUri)
          sv->setCheckState(0, Qt::Unchecked);
      }
      dirty = true;
    }
  }
}

void te::qt::plugins::wtss::WtssDialog::onHideButtonClicked()
{
  if(m_ui->serverFrame->isVisible())
  {
    m_ui->serverFrame->hide();
    m_ui->m_hideToolButton->setIcon(QIcon::fromTheme("show"));
  }
  else
  {
    m_ui->serverFrame->show();
    m_ui->m_hideToolButton->setIcon(QIcon::fromTheme("hide"));
  }
}

void te::qt::plugins::wtss::WtssDialog::onHideCoordSelectedsClicked()
{
  if(m_ui->coordSframe->isVisible())
  {
    m_ui->coordSframe->hide();
    m_ui->m_hideCoordSToolButton->setIcon(QIcon::fromTheme("show"));
  }
  else
  {
    m_ui->coordSframe->show();
    m_ui->m_hideCoordSToolButton->setIcon(QIcon::fromTheme("hide"));
  }
}

void te::qt::plugins::wtss::WtssDialog::onHelpButtonClicked()
{
}

void te::qt::plugins::wtss::WtssDialog::onCloseButtonClicked()
{
  clearCanvas();
  this->close();
}

void te::qt::plugins::wtss::WtssDialog::onImportGraphClicked()
{
  QString csvFile =
      QFileDialog::getOpenFileName(this, "Select a TimeSeries CSV file",
                                   QDir::currentPath(), "CSV file (*.csv)");

  if(csvFile.isEmpty())
    return;

  QFileInfo file(csvFile);
  std::ifstream csv(csvFile.toUtf8().data());

  std::string str;
  std::getline(csv, str);

  m_timeSeriesVec.clear();

  m_timeSeriesVec.push_back(
      new te::st::TimeSeries(file.baseName().toUtf8().data()));

  while(std::getline(csv, str))
  {
    QString qstr = QString::fromUtf8(str.c_str());
    QStringList fields = qstr.split(',');
    QStringList date = fields.front().split("/");

    te::dt::DateTime *time = new te::dt::Date(boost::gregorian::date(
        date.at(2).toInt(), date.at(1).toInt(), date.at(0).toInt()));

    double value = fields.back().toDouble();

    m_timeSeriesVec.back()->add(time, value);
  }
  plotTimeSeries();
}

void te::qt::plugins::wtss::WtssDialog::onExportGraphClicked()
{
  if(m_result.query.attributes.empty())
    return;

  QString cvName(QString::fromUtf8(m_result.coverage.name.c_str()));

  std::vector<cxx::queried_attribute_t> attributes =
      m_result.coverage.queried_attributes;

  for(unsigned int i = 0; i < attributes.size(); i++)
  {
    cxx::queried_attribute_t attribute = attributes[i];

    QString csvFile = QFileDialog::getSaveFileName(
        this, tr("Save File"),
        QDir::currentPath() + "/" + attribute.name.c_str() + ".csv",
        "CSV file (*.csv)");

    if(csvFile.isEmpty())
      return;

    QFileInfo info(csvFile);

    if(info.suffix().isEmpty())
      csvFile.append(".csv");

    std::ofstream myfile(csvFile.toUtf8().data());

    myfile << "Timeline,Value" << std::endl;

    for(unsigned int j = 0; j < attribute.values.size(); ++j)
    {
      QJsonObject j_attribute =
          te::qt::plugins::wtss::ServerManager::getInstance().getAttribute(
              QString::fromUtf8(m_lastQueriedServer.c_str()), cvName,
              QString::fromUtf8(attribute.name.c_str()));

      cxx::date d = m_result.coverage.timeline[j];

      double value =
          attribute.values[j] * j_attribute["scale_factor"].toDouble();

      myfile << d.day << "/" << d.month << "/" << d.year << "," << value
             << std::endl;
    }
    myfile.close();
  }
}

void te::qt::plugins::wtss::WtssDialog::onQueryButtonClicked()
{
  cxx::timeseries_query_t query;

  if(m_ui->latLineEdit->text().isEmpty() &&
     m_ui->longLineEdit->text().isEmpty())
  {
    QMessageBox::warning(this, tr("Web Time Series"),
                         tr("Input the latitude and longitude coordinate."));
    return;
  }
  else if(m_ui->latLineEdit->text().isEmpty())
  {
    QMessageBox::warning(this, tr("Web Time Series"),
                         tr("Input the latitude coordinate."));
    return;
  }
  else if(m_ui->longLineEdit->text().isEmpty())
  {
    QMessageBox::warning(this, tr("Web Time Series"),
                         tr("Input the longitude coordinate."));
    return;
  }

  query.latitude = m_ui->latLineEdit->text().toDouble();
  query.longitude = m_ui->longLineEdit->text().toDouble();

  QDate startDate = m_ui->m_startDateEdit->date();
  QDate endDate = m_ui->m_endDateEdit->date();

  if(!startDate.operator<=(endDate))
  {
    QMessageBox::warning(this, tr("Web Time Series"),
                         tr("The date informed is invalid."));
    return;
  }

  te::qt::plugins::wtss::ServerManager::getInstance().addDateFilter(
      startDate.toString("dd/MM/yyyy"), endDate.toString("dd/MM/yyyy"));

  doTimeseriesQuery(query);
}

void te::qt::plugins::wtss::WtssDialog::onAddCoordToList(
    QListWidgetItem *coordSelected)
{
  cxx::timeseries_query_t query;

  QStringList coord = coordSelected->text().split(",");

  query.longitude = coord.front().toDouble();
  query.latitude = coord.back().toDouble();

  doTimeseriesQuery(query);

  addMarker(query.longitude, query.latitude);
}

void te::qt::plugins::wtss::WtssDialog::onGetPointCoordinate(QPointF &coord)
{
  std::auto_ptr<te::srs::Converter> converter(new te::srs::Converter());

  try
  {
    converter->setSourceSRID(m_mapDisplay->getSRID());
    converter->setTargetSRID(4326);
  }
  catch(...)
  {
    QMessageBox::warning(this, tr("Web Time Series"), tr("The display SRS is unknown."));
    return;
  }

  te::gm::Coord2D c;

  converter->convert(coord.x(), coord.y(), c.x, c.y);

  cxx::timeseries_query_t query;

  query.longitude = c.x;
  query.latitude = c.y;

  if(m_ui->m_serverTreeWidget->topLevelItemCount() > 0)
  {
    doTimeseriesQuery(query);

    addMarker(query.longitude, query.latitude);
  }

  WtssDialog::setModal(false);
  WtssDialog::show();
}

void te::qt::plugins::wtss::WtssDialog::loadSettings()
{
  try
  {
    j_config = te::qt::plugins::wtss::ServerManager::getInstance()
                   .loadSettings()
                   .object();

    serverSettings();

    dateSettings();
  }
  catch(...)
  {
    QMessageBox::warning(this, tr("Web Time Series"),
                         tr("An error has occurred, "
                            "when try to load settings."));
  }
}

void te::qt::plugins::wtss::WtssDialog::serverSettings()
{
  QJsonObject j_servers = j_config.find("servers").value().toObject();

  if(j_servers.isEmpty())
    return;

  for(QJsonObject::iterator it = j_servers.begin(); it != j_servers.end(); ++it)
    addServer(it.key());
}

void te::qt::plugins::wtss::WtssDialog::dateSettings()
{
  QJsonObject j_datefilter = j_config.find("date_filter").value().toObject();

  if(j_datefilter.isEmpty())
    return;

  QString startDate = j_datefilter.find("start_date").value().toString();

  QString endDate = j_datefilter.find("end_date").value().toString();

  m_ui->m_startDateEdit->setDate(QDate::fromString(startDate, "dd/MM/yyyy"));

  m_ui->m_endDateEdit->setDate(QDate::fromString(endDate, "dd/MM/yyyy"));
}

void te::qt::plugins::wtss::WtssDialog::addServer(QString server)
{
  QJsonObject j_servers = j_config.find("servers").value().toObject();

  QJsonObject j_server = j_servers.find(server).value().toObject();

  QTreeWidgetItem *serverItem =
      new QTreeWidgetItem(m_ui->m_serverTreeWidget, treeItemType::server);

  serverItem->setText(0, server);

  serverItem->setFlags(serverItem->flags() | Qt::ItemIsUserCheckable);

  bool active = j_server.find("active").value().toBool();

  if(active)
    serverItem->setCheckState(0, Qt::Checked);
  else
    serverItem->setCheckState(0, Qt::Unchecked);

  addCoverage(serverItem, j_server);
}

void te::qt::plugins::wtss::WtssDialog::addCoverage(
    QTreeWidgetItem *serverItem, QJsonObject j_server)
{
  QJsonObject j_coverages = j_server["coverages"].toObject();

  for(QJsonObject::iterator it = j_coverages.begin(); it != j_coverages.end();
      it++)
  {
    bool active = it.value().toObject().find("active").value().toBool();
    QTreeWidgetItem *coverageItem =
        new QTreeWidgetItem(treeItemType::coverage);

    coverageItem->setText(0, it.key());

    coverageItem->setFlags(coverageItem->flags() | Qt::ItemIsUserCheckable);

    if(active)
      coverageItem->setCheckState(0, Qt::Checked);
    else
      coverageItem->setCheckState(0, Qt::Unchecked);

    serverItem->addChild(coverageItem);

    addAtributes(coverageItem, it.value().toObject());
  }
}

void te::qt::plugins::wtss::WtssDialog::addAtributes(
    QTreeWidgetItem *coverageItem, QJsonObject j_coverage)
{
  QJsonObject j_attributes = j_coverage.find("attributes").value().toObject();

  for(QJsonObject::iterator it = j_attributes.begin(); it != j_attributes.end();
      it++)
  {
    QString attribute = it.key();

    bool active = it.value().toObject().find("active").value().toBool();
    QTreeWidgetItem *attributeItem =
        new QTreeWidgetItem(treeItemType::attribute);

    attributeItem->setText(0, attribute);
    attributeItem->setFlags(attributeItem->flags() | Qt::ItemIsUserCheckable);

    if(active)
      attributeItem->setCheckState(0, Qt::Checked);
    else
      attributeItem->setCheckState(0, Qt::Unchecked);

    coverageItem->addChild(attributeItem);
  }
}

bool te::qt::plugins::wtss::WtssDialog::validateQuery()
{
  if(m_checkServer)
  {
    if(m_checkCoverage)
    {
      if(m_checkAttribute)
        return true;
      else
      {
        QMessageBox::warning(this, tr("Web Time Series"),
                             tr("Please, select an attribute."));
        return false;
      }
    }
    else
    {
      QMessageBox::warning(this, tr("Web Time Series"),
                           tr("Please, select a coverage."));
      return false;
    }
  }
  else
  {
    QMessageBox::warning(this, tr("Web Time Series"),
                         tr("Please, select a server"));
    return false;
  }
}

void te::qt::plugins::wtss::WtssDialog::doTimeseriesQuery(
    cxx::timeseries_query_t query)
{
  clearCanvas();

  m_ui->latLineEdit->setText(QString::number(query.latitude));
  m_ui->longLineEdit->setText(QString::number(query.longitude));

  te::qt::widgets::ScopedCursor c(Qt::WaitCursor);

  QJsonObject j_object = te::qt::plugins::wtss::ServerManager::getInstance()
                             .loadSettings()
                             .object();

  if(j_object.isEmpty())
    return;

  QJsonObject j_servers = j_object.find("servers").value().toObject();

  QJsonObject j_server;

  m_checkServer = false;
  m_checkCoverage = false;
  m_checkAttribute = false;

  for(QJsonObject::iterator it_server = j_servers.begin();
      it_server != j_servers.end(); ++it_server)
  {
    j_server = it_server.value().toObject();

    if(j_server.find("active").value().toBool())
    {
      std::string serverUri = it_server.key().toStdString();

      QJsonObject j_coverages = j_server["coverages"].toObject();

      for(QJsonObject::iterator it_coverages = j_coverages.begin();
          it_coverages != j_coverages.end(); it_coverages++)
      {
        if(it_coverages.value().toObject().find("active").value().toBool())
        {
          cxx::timeseries_query_t new_q;

          new_q.latitude = query.latitude;
          new_q.longitude = query.longitude;
          new_q.coverage_name = it_coverages.key().toUtf8().data();

          QJsonObject j_attributes = it_coverages.value()
                                         .toObject()
                                         .find("attributes")
                                         .value()
                                         .toObject();

          for(QJsonObject::iterator it_attributes = j_attributes.begin();
              it_attributes != j_attributes.end(); it_attributes++)
          {
            QJsonObject j_attribute = it_attributes.value().toObject();

            if(j_attribute["active"].toBool() == true)
            {
              new_q.attributes.push_back(it_attributes.key().toUtf8().data());
              m_checkAttribute = true;
            }
          }

          if(new_q.attributes.size() > 0)
          {
            cxx::client remote(serverUri);

            try
            {
              m_result = remote.time_series(new_q);
              m_lastQueriedServer = it_server.key().toUtf8().data();
              convertToTimeSeries(m_result);
            }
            catch(const std::exception &e)
            {
              QMessageBox::warning(this, tr("Web Time Series"),
                                   tr("The coordinates "
                                      "informed are invalid."));
              return;
            }
          }
          m_checkCoverage = true;
        }
      }
      m_checkServer = true;
    }
  }

  if(!validateQuery())
  {
    m_chartDisplay->detachItems(te::qt::widgets::TIMESERIES_CHART);
    return;
  }

  addLocation(query.longitude, query.latitude);

  plotTimeSeries();
}

void te::qt::plugins::wtss::WtssDialog::convertToTimeSeries(
    cxx::timeseries_query_result_t result)
{
  QString cvName(QString::fromUtf8(result.coverage.name.c_str()));

  std::vector<cxx::queried_attribute_t> attributes =
      result.coverage.queried_attributes;

  m_timeSeriesVec.clear();

  for(unsigned int i = 0; i < attributes.size(); i++)
  {
    cxx::queried_attribute_t attribute = attributes[i];

    m_timeSeriesVec.push_back(new te::st::TimeSeries(attribute.name));

    for(unsigned int j = 0; j < attribute.values.size(); ++j)
    {
      QJsonObject j_attribute =
          te::qt::plugins::wtss::ServerManager::getInstance().getAttribute(
              QString::fromUtf8(m_lastQueriedServer.c_str()), cvName,
              QString::fromUtf8(attribute.name.c_str()));

      cxx::date d = result.coverage.timeline[j];
      te::dt::DateTime *time =
          new te::dt::Date(boost::gregorian::date(d.year, d.month, d.day));

      double value =
          attribute.values[j] * j_attribute["scale_factor"].toDouble();

      m_timeSeriesVec.at(i)->add(time, value);
    }
  }
}

std::vector<te::st::TimeSeries *>
te::qt::plugins::wtss::WtssDialog::getTimeSeries()
{
  return m_timeSeriesVec;
}

void te::qt::plugins::wtss::WtssDialog::plotTimeSeries()
{
  m_chartDisplay->detachItems(te::qt::widgets::TIMESERIES_CHART);

  if(getTimeSeries().empty())
    return;

  for(unsigned int i = 0; i < getTimeSeries().size(); ++i)
  {
    m_timeSeriesChart =
        new te::qt::widgets::TimeSeriesChart(getTimeSeries().at(i));

    m_timeSeriesChart->setPen(QPen(randomColor()));
    m_timeSeriesChart->attach(m_chartDisplay);
    m_timeSeriesChart->setTitle(tr(getTimeSeries().at(i)->getId().c_str()));

    m_chartDisplay->insertLegend(new QwtLegend(), QwtPlot::RightLegend);
    m_chartDisplay->adjustDisplay();
    m_chartDisplay->show();
  }
}

void te::qt::plugins::wtss::WtssDialog::defineDisplay()
{
  m_chartStyle = new te::qt::widgets::ChartStyle();
  m_chartStyle->setTitle(QString::fromUtf8("Web Time Series"));
  m_chartStyle->setAxisX(QString::fromUtf8("Timeline"));
  m_chartStyle->setAxisY(QString::fromUtf8("Value"));

  m_chartDisplay = new te::qt::widgets::ChartDisplay(
      m_ui->m_timeSeriesFrame, QString::fromUtf8("Web Time Series"),
      m_chartStyle);

  QGridLayout *m_layout = new QGridLayout(m_ui->m_timeSeriesFrame);
  m_layout->addWidget(m_chartDisplay, 0, 0);
  m_layout->setContentsMargins(0, 0, 0, 0);
}

void te::qt::plugins::wtss::WtssDialog::defineMarker()
{
  te::se::Stroke *stroke = te::se::CreateStroke("#000000", "1");
  te::se::Fill *fill = te::se::CreateFill("#000000", "1.0");
  m_marker = te::se::CreateMark("cross", stroke, fill);
  m_rgbaMarker =
      te::map::MarkRendererManager::getInstance().render(m_marker, 12);
}

void te::qt::plugins::wtss::WtssDialog::setPointPicked(QPointF &coord)
{
  emit pointPicked(coord);
}

void te::qt::plugins::wtss::WtssDialog::addMarker(double x, double y)
{
  m_mapDisplay->getDisplayPixmap()->fill(Qt::transparent);

  const te::gm::Envelope &env = m_mapDisplay->getExtent();

  te::qt::widgets::Canvas canvas(m_mapDisplay->getDraftPixmap());

  canvas.setWindow(env.m_llx, env.m_lly, env.m_urx, env.m_ury);
  canvas.setPointPattern(m_rgbaMarker, 12, 12);

  std::auto_ptr<te::srs::Converter> converter(new te::srs::Converter());

  converter->setSourceSRID(4326);
  converter->setTargetSRID(m_mapDisplay->getSRID());

  te::gm::Coord2D coord;

  converter->convert(x, y, coord.x, coord.y);

  te::gm::Point point(coord.x, coord.y);

  canvas.setPointColor(te::color::RGBAColor(0, 0, 0, 0));
  canvas.draw(&point);

  m_mapDisplay->repaint();
}

QColor te::qt::plugins::wtss::WtssDialog::randomColor()
{
  int r = rand() % 255;
  int g = rand() % 255;
  int b = rand() % 255;
  return QColor(r, g, b);
}

void te::qt::plugins::wtss::WtssDialog::addLocation(double x, double y)
{
  QList<QListWidgetItem *> coord = m_ui->m_coordSListWidget->findItems(
      QString::number(x) + "," + QString::number(y), Qt::MatchExactly);
  if(coord.length() != 0)
    return;

  m_ui->m_coordSListWidget->addItem(QString::number(x) + "," +
                                    QString::number(y));
}

void te::qt::plugins::wtss::WtssDialog::clearCanvas()
{
  te::qt::widgets::Canvas canvas(m_mapDisplay->getDraftPixmap());

  canvas.clear();

  m_mapDisplay->refresh();
}

void te::qt::plugins::wtss::WtssDialog::closeEvent(QCloseEvent *e)
{
  m_chartDisplay->detachItems(te::qt::widgets::TIMESERIES_CHART);

  m_ui->m_coordSListWidget->clear();

  clearCanvas();

  emit close();
}
