/*
  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

  This file is part of the TerraLib - a Framework for building GIS enabled
  applications.

  TerraLib is free software: you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation, either version 3 of the License,
  or (at your option) any later version.

  TerraLib is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with TerraLib. See COPYING. If not, write to
  TerraLib Team at <terralib-team@terralib.org>.
  */

/*!
  \file terralib/vm/python/VirtualMachine.cpp

  \brief Implementation of a virtual machine for the Python Programming
  Language.

  \author Frederico Augusto Bedê
  */

// TerraLib
#include "VirtualMachine.h"
#include "../../core/filesystem/FileSystem.h"
#include "../../core/translator/Translator.h"
#include "../../core/utils/Platform.h"
#include "../core/Exception.h"

// Python
#include <Python.h>

// STL
#include <functional>

#include <iostream>

void PythonPrinting(std::string filename)
{
  std::string stdOutErr =
      "import sys\n\
class CatchOut:\n\
  def __init__(self):\n\
    self.value = ''\n\
  def write(self, txt):\n\
    self.value += txt\n\
catchOut = CatchOut()\n\
sys.stdout = catchOut\n\
sys.stderr = catchOut\n\
";  // this is python code to redirect stdouts/stderr

  PyObject *pModule = PyImport_AddModule("__main__");  // create main module
  PyRun_SimpleString(stdOutErr.c_str());  // invoke code to redirect
  //  PyRun_SimpleString("print('hello world')");
  FILE *py_file;
  py_file = fopen(filename.c_str(), "r");
  PyRun_SimpleFile(py_file, filename.c_str());
  PyObject *catcher = PyObject_GetAttrString(pModule, "catchOut");
  fclose(py_file);

  PyObject *output = PyObject_GetAttrString(catcher, "value");
  std::string out = PyString_AsString(output);
  std::cout << out;

  Py_DECREF(output);
  Py_DECREF(catcher);
}

te::vm::python::VirtualMachine::VirtualMachine()
{
  Py_Initialize();
}

te::vm::python::VirtualMachine::~VirtualMachine()
{
  Py_Finalize();
}

std::string te::vm::python::VirtualMachine::getName() const
{
  return "Python";
}

std::string te::vm::python::VirtualMachine::getTitle() const
{
  std::string msg(
      TE_TR("TerraLib Virtual Machine for the Python Programming Language"));

  return msg;
}

std::string te::vm::python::VirtualMachine::getDescription() const
{
  std::string msg(TE_TR(
      "This Virtual Machine allows applications to execute python scripts."
      "\nAll the TerraLib API will be available for TerraLib programmers "
      "in the Python environment"));

  return msg;
}

void te::vm::python::VirtualMachine::build(const std::string &file)
{
  m_filename = file;
}

void te::vm::python::VirtualMachine::execute()
{
  PythonPrinting(m_filename);
}

void te::vm::python::VirtualMachine::pause()
{
}

void te::vm::python::VirtualMachine::stop()
{
}

void te::vm::python::VirtualMachine::reset()
{
}
